#ifndef FONT_H
#define FONT_H
#include "config.h"
/*
  ・font ID (fid):
        1文字分のフォントの識別子(フォントID)。列挙型font_id_tで表され、(たぶん)1ワード。
        本ライブラリで用意した文字のみに通し番号および列挙子を割り当てる。
        したがって、用意された文字は列挙子を用いて指定可能だが、
        用意されていない文字は扱えない。
        また、ASCIIやEUCなどの文字コードとも一致しないので、
        char型の文字データからは変換が必要。
  ・font sequence (fontseq):
        0個以上のfont_id_t型の並びで、末尾の値がfi_endであるもの。
        フォントIDの並びであり、一つの文字列に対応する。
        Cでchar*により文字列を扱う方法に準拠しているので、
        font_id_t*で一つのfontseqを表し、
        fontseqを扱う関数はこのポインタを受け取って、
        fi_endでターミネートされた文字(フォント)の並びとして処理する。
        alloc_fontseq()を使うと、文字列をfontseqに変換したポインタを取得でき、
        fontseqの実体領域は自動確保される。
   ・fontseq array
        fontseqの配列。複数の文字列をまとめて扱う場合に使用。
        
*/

#define FONTSIZE 16

typedef enum {
    //    fi_invalid = -1,
    fi_begin = 0,
    fi_tofu = 0,
    fi_space,
    fi_kuten,
    fi_toten,
	fi_nakaguro,
	fi_question,
	fi_bikkuri,
    fi_dash,
	fi_ldots,
	fi_lparen,
	fi_rparen,
	fi_lbracket,
	fi_rbracket,
	fi_lkagikakko,
	fi_rkagikakko,
    fi_a,
    fi_i,
    fi_u,
    fi_e,
    fi_o,
    fi_la,
    fi_li,
    fi_lu,
    fi_le,
    fi_lo,
    fi_ka,
    fi_ki,
    fi_ku,
    fi_ke,
    fi_ko,
    fi_ga,
    fi_gi,
    fi_gu,
    fi_ge,
    fi_go,
    fi_sa,
    fi_si,
    fi_su,
    fi_se,
    fi_so,
    fi_za,
    fi_zi,
    fi_zu,
    fi_ze,
    fi_zo,
    fi_ta,
    fi_ti,
    fi_tu,
    fi_te,
    fi_to,
    fi_ltu,
    fi_da,
    fi_di,
    fi_du,
    fi_de,
    fi_do,
    fi_na,
    fi_ni,
    fi_nu,
    fi_ne,
    fi_no,
    fi_ha,
    fi_hi,
    fi_hu,
    fi_he,
    fi_ho,
    fi_ba,
    fi_bi,
    fi_bu,
    fi_be,
    fi_bo,
    fi_pa,
    fi_pi,
    fi_pu,
    fi_pe,
    fi_po,
    fi_ma,
    fi_mi,
    fi_mu,
    fi_me,
    fi_mo,
    fi_ya,
    fi_yu,
    fi_yo,
    fi_lya,
    fi_lyu,
    fi_lyo,
    fi_ra,
    fi_ri,
    fi_ru,
    fi_re,
    fi_ro,
    fi_wa,
    fi_wo,
    fi_n,

    fi_n0,
    fi_n1,
    fi_n2,
    fi_n3,
    fi_n4,
    fi_n5,
    fi_n6,
    fi_n7,
    fi_n8,
    fi_n9,

    fi_al_c,
    fi_al_f,
    fi_al_o,
    fi_al_n,
    fi_al_s,

    fi_ky_g,
    fi_ky_p,
    fi_ky_r,
    fi_ky_q,

	fi_ky_leftarrow,
	fi_ky_rightarrow,
	fi_ky_uparrow,
	fi_ky_downarrow,
	fi_ky_spacel,
	fi_ky_spacer,
	fi_ky_esc,
	fi_ky_enter,

	fi_time,
	fi_walk,
	fi_jump,
	fi_air,
	fi_lift,
	fi_blast,
	fi_sting,
	fi_bone,
	fi_smash,
	fi_flame,
	fi_goal,
	fi_gondin,
	fi_gondin_lost,
	fi_chee,
	fi_hatopo,
	fi_yaramasi,
	fi_gagga,
	fi_fwall,
	fi_iwall,
    fi_switch,
	fi_teleporter,
	fi_sue,
	fi_ojigiri,
	fi_chu,
	fi_jumeme,
	fi_siguku,
	fi_mingo,

	fi_plus,
	fi_multiply,
    fi_slash,
	fi_equal,
    fi_op_le,
    fi_op_ge,
    fi_op_lt,
    fi_op_gt,
	fi_checkbox,
	fi_check,
	fi_check_weak,
	fi_star_off,
	fi_star_on,

    fi_end
} font_id_t;

// [FontName]形式でフォントを指定するためのテーブルエントリ
typedef struct {
    char *font_name;
    font_id_t fid;
} font_name_t;

typedef struct {
    font_id_t fid; // フォントの識別子(1～)
    char *mbchr;   // そのフォントに対応する文字(マルチバイト文字に対応するため文字列)
    char **xpm;    // フォント画像のxpmデータ
} font_info_t;

// fidから変換したfixの最大値
// テーブルfix2fidはこのサイズで確保する
// fixは
// ・fidと一対一対応する必要があるが、連続している必要はない
// ・jcodeが返すマルチバイト文字コードcodeからの変換計算は軽い方が良いが、
//   stringからfix経由のfid変換はなるべく事前に行うことを想定しているので、
//   コンスタントオーバヘッドはそれほど気にしなくて良い
// ・将来的に一部の文字についてfix値を変更しても、fidとの対応関係は
//   font_info_list経由で自動計算されるので問題ない
#define FIX_MAX (1024*1024) 

// [FontName]で使うフォント名の個数上限
#define FONT_NAME_NUM_MAX 1024

// 実行中に変更しない文字列リテラルについては、
// alloc_fontseqに文字列を渡すことで、fontseqへのポインタに変換できる。
// この場合、fontseqの実体はバッファに自動確保される。
// バッファは次のマクロ値バイトが自動確保され、
// あふれる度に同サイズが追加確保される
#define FONTSEQ_BUF_SIZE 64*1024


#define FONTSEQ_ARRAY_BUF_SIZE 64
#define FONTSEQ_ARRAY_PACK_LIMIT (FONTSEQ_ARRAY_BUF_SIZE/2)

void font_init(void);
void draw_font(int x, int y, font_id_t fid);
void draw_fontseq(int x, int y, font_id_t *fid);
int mbchr2fix(char **chr);
int str2fontseq(char *str, font_id_t fid[]);
font_id_t *alloc_fontseq(char *str);
font_id_t **alloc_fontseq_array(int size);
int fontseq_cpy(font_id_t *dest, font_id_t *src);
int fontseq_len(font_id_t *fontseq);
int fontseq_cat(font_id_t *dest, font_id_t *src);
int uint2fontseq(unsigned int n, int len, font_id_t fid[]);
int inc_fontseq(int len, font_id_t fid[]);
int dec_fontseq(int len, font_id_t fid[]);

#endif
