#ifndef GAME_H
#define GAME_H
#include "config.h"
#include "param.h"
#include "record.h"
#include "objective.h"
#include "font.h"

typedef enum {
	st_notyet, // 未達成
    st_sofar,  // 現状達成だが未確定
	st_done,   // 達成済み
	st_failed  // 失敗(今後の達成が不可能)
} objective_stat_t;

typedef struct {
	objective_operator_t op;
	unsigned int val;
	objective_stat_t stat;
	font_id_t stat_fontseq[2];
} game_objective_t;

typedef struct {
    unsigned int record[re_end];
	// objectives[]アクセス用
	// (ix=objective_ix_table[re_xxx])==-1ならre_xxxはobjectiveではない
	// ix=0..OBJECTIVE_NUM-1なら、get_objective(ix)で
	// re_xxxに対応するobjectiveが得られる
	// また、objective_op[ix], objective_val[ix]は
	// get_objectiveで得られる比較演算子や比較値のキャッシュである
	int objective_ix_table[re_end];
	game_objective_t objectives[OBJECTIVE_NUM];
	
//    unsigned int time;
	unsigned int gondin_toeat;
	unsigned int gondin_total;
    int frames_per_sec;
    unsigned int score_time;
    unsigned int score_gondin;
//    unsigned int score_total;
#ifdef DEBUG
    // デバッグ用に当たり判定を可視化するかどうか
    int show_hitmarker;
#endif
} game_t;

#define GameRecord(re) (game.record[re])
#define SetGameRecord(re, v) (game.record[re]=v)
#define IncGameRecord(re) (game.record[re]++, update_objective_stat_inc(re))
#define AddGameRecord(re, v) ((game.record[re] += v), update_objective_stat_inc(re))
#define DecGameRecord(re) (game.record[re]--, update_objective_stat_dec(re))

extern game_t game;
extern int jump_key_released; // ジャンプキーを放したら羽ばたきを止めるためとりあえず公開
extern int has_space_released; // gameendまでスペースが残るのを防ぐためとりあえず。

#define GetFramesPerSec() (game.frames_per_sec)
#define GetShowHitMarker() (game.show_hitmarker)
#define GetTime() (GameRecord(re_time))
#define SetTime(t) (SetGameRecord(re_time, t))
#define DecTime() (DecGameRecord(re_time))
#define GetScoreTime() (game.score_time)
#define GetScoreGondin() (game.score_gondin)
#define GetGondinEaten() (GameRecord(re_gondin_eaten))
#define SetGondinEaten(g) (SetGameRecord(re_gondin_eaten, g))
#define IncGondinEaten() (IncGameRecord(re_gondin_eaten))
#define GetScoreTotal() (GameRecord(re_score]))

void game_init(void);
void setup_game(void);
void game_keyh(int key);
void game_key_releaseh(int key);
void game_timerh(void);
void pause_timerh(void);
int inc_gondin_eaten(void);
void quit_game(void);
void retry_game(void);
int update_objective_stat_inc(record_entry_t re);
int update_objective_stat_dec(record_entry_t re);
game_objective_t *get_game_objective(int no);
void finalize_objective_stat(void);
void inc_kill_record(obj_type_t type);

#endif




