#ifndef OBJ_H
#define OBJ_H
#include "config.h"
#include "uty.h"
#include "param.h"

#define Ltx(obj) ((obj)->x)
#define Lty(obj) ((obj)->y)
#define Rtx(obj) ((obj)->x+(obj)->width-1)
#define Rty(obj) ((obj)->y)
#define Lbx(obj) ((obj)->x)
#define Lbpx(obj) ((obj)->prev_x)
#define Lby(obj) ((obj)->y+(obj)->height-1)
#define Rbx(obj) ((obj)->x+(obj)->width-1)
#define Rbpx(obj) ((obj)->prev_x+(obj)->width-1)
#define Rby(obj) ((obj)->y+(obj)->height-1)
#define GridX(obj) (Ltx(obj) >> GRID_SCALE)
#define GridY(obj) (Lty(obj) >> GRID_SCALE)
#define ObjName(obj) (get_type_name((obj)->type)

typedef enum {
    oc_begin = 0,
    oc_static = 0,
    oc_hard_dynamic,
    oc_soft_dynamic,
    oc_end
} obj_class_t;

#define OS_ALIVE (1 << 13)
#define OS_STUN (1 << 12)

typedef enum {
    os_invalid = 0,
	os_retry,    // Cheeに設定するとステージリスタート
	os_quit,     // cheeに設定するとステージセレクトに戻る
    os_ghost,    // Cheeが幽霊
	os_gameover, // ゲームオーバー
	os_exit,
	os_clear, // Cheeがステージクリア
	os_show_result, // ステージクリアメッセージ表示中
    os_smash_h1, // 水平方向に1単位潰れた状態。次ステップでbangに移行
    os_smash_h2, // 水平方向に2単位潰れた状態。次ステップでbangに移行
    os_smash_v1, // 垂直方向に1単位潰れた状態。次ステップでbangに移行
    os_smash_v2, // 水平方向に2単位潰れた状態。次ステップでbangに移行
    os_bang,     // 破裂グラフィック。遷移元は複数あるが次ステップでos_dead
    os_pierced,  // とげ・ごんでぃんに刺さった
    os_burned,   // 炎で焦げた
	os_electrocuted, // バリアで感電した
    os_eaten,    // 食べられた。後の処理(状態遷移)はオブジェクト依存
    os_digested, // 食べられてしまい消化中。
	os_dead,     // 死亡しオブジェクト削除。kill_recordのためburnedやbangなどからここに遷移
    os_req_on,
    os_req_off,
    os_on,
    os_off,
//    os_pushed_up, // リフトがスイッチを1単位押し上げるのに使う
	os_weighted_on,
	os_weighted_on_int,
	os_weighted_off,
	os_weighted_off_int,
    os_nonweighted_on,
//    os_nonweighted_on_int,
    os_nonweighted_off,
    os_goal_req_open, // ゴールを開き始める
	os_goal_opening,
	os_goal_opened,
	os_goal_closing,
	os_gondin_eaten,
    os_nya_halting,
    
    os_alive_begin = OS_ALIVE,

    os_none,     // 通常の生存状態 
    os_digesting,// おじぎり草が消化中。
    os_vomitting,// おじぎり草が骨を吐き出し中
    os_eating,   // 何か食べている(おじぎり草、ちゅっちゅ)
	os_blast, // Gaggaが攻撃状態
	os_head_high,    // すーの首上下用
	os_head_low,
	os_breath_high, // すーの攻撃状態
	os_breath_low,
	os_standing,
    os_cancel,   // 行動中止。とりあえずおじぎり草の捕食中止に使う
    os_siguku_landed, // しぐくが接地(氷や斜面含む)
    os_siguku_splashing, // しぐくが破裂
    os_siguku_freezing,
	os_jumeme_melting_upper,
	os_jumeme_melting_side,
	os_jumeme_melting_lower,
    os_mingo_attack, // os_attackとしてos_blastと統合できる?
    
	os_blasted = OS_ALIVE + OS_STUN, // 吹っ飛ばされ中
    os_teleporting,
    os_chee_eating, // ちぃちぃがごんでぃんを食べている。操作不能にするためSTUN属性

} obj_state_t;

typedef enum {
    er_none = 0,
    er_smash_h1 = os_smash_h1,
    er_smash_h2 = os_smash_h2,
    er_smash_v1 = os_smash_v1,
    er_smash_v2 = os_smash_v2,
    er_excluded_horizontal = 0x100,
    er_excluded_vertical = 0x200,
} exclude_result_t;

#define IsSmashed(re) ((re)&0xff)
#define IsExcludedHorizontal(re) ((re)&0x100)
#define IsExcludedVertical(re) ((re)&0x200)
#define IsExcluded(re) ((re)&0x300)
#define ExcludeResult2SmashState(re) ((re)&0xff)

typedef enum {
//    at_immune_ojigiri, // おじぎり草に食べられない
	at_meat,           // 動物(おじぎり草の捕食対象)
	at_vegetable,      // 植物(おじぎり草は反応しないが、がっがは殴る)
	at_item,           // しぐく、(!! 将来的に骨?)

    at_rooted,         // 根を下ろしている(左右からexcludeされたら即死) 
	at_can_climb,    // 階段を登れる
	at_immune_pierce,   // 棘無効
	at_immune_flame,    // 炎無効
	at_immune_water,    // 水無効
	at_turn_to_mist,    // 火や水にやられたとき霧状になる(重力無効になる)
} attr_bit_t;

#define  SetAttr(obj, at)                        \
     ((obj)->attr |= (1 << (at)))
#define CheckAttr(obj, at)                 \
     ((obj)->attr & (1 << (at)))

typedef enum {
    ot_begin = 0, ot_none = 0,

	ot_gondin,   // ゴンディンの実

	ot_hatopo,   // はとぽ
	ot_gagga,    // がっが
	ot_yaramasi, // 
    ot_ojigiri,  // おじぎり草
	ot_sue,      // すー
	ot_chu,      // ちゅっちゅ
    ot_mingo,    // みんご
	ot_jumeme,   // じゅめめ
	ot_siguku,   // しぐく
    ot_chee,     // プレイヤーキャラ
    ot_creature_end,
	ot_blank = ot_creature_end,    // 空白
	ot_teleport_dest, // テレポート先(実際には空白となりオブジェクトとしては生成されない)
	ot_teleporter,
	ot_pipe,
	ot_wall,     // 破壊不能・移動・落下しない壁
    ot_fwall,    // 破壊可能な壁
	ot_iwall,    // 氷の壁
	ot_barrier,  // バリア
	ot_goal,
    ot_slope_l,  // 斜面左・右
    ot_slope_r,
    ot_stairs_l,  // 階段左・右
    ot_stairs_r,
    ot_rstairs_l,  // 逆さ階段左・右
    ot_rstairs_r,
	ot_sign,
	ot_sign_hang,




	ot_lift,     // エレベータ
	ot_mswitch,  // モーメンタリスイッチ(乗っている間だけON)
	ot_aswitch,  // オルタネートスイッチ(乗る度にON/OFF切り換え)

    ot_nya,      // 
	ot_punch,    // 
	ot_flame,

    ot_sting,    // 破壊不能・移動不能な棘
                 // ※このタイプはcreate関数を呼ぶときのみ使われ、
                 //   実際には指定した方向に応じ以下のいずれかが生成される
                 //   mapファイルには単一シンボル+方向指定で記述したい一方、
                 //   方向毎に描画や当たり判定が異なり
                 //   実行中はそれぞれ別タイプとした方が効率がよいためである。
    ot_sting_l,  // 左向きの棘
    ot_sting_r,  // 右向きの棘
    ot_sting_u,  // 上向きの棘
    ot_sting_d,  // 下向きの棘


	ot_hub,
//    ot_mvfloor,  // 動く床
    ot_end
} obj_type_t;


typedef struct {
    int move_x;         // 水平方向の移動量
    int land_ofst;
    int inertia_x;      // リフトの慣性(ジャンプ開始時にjump_xに代入)
    int inertia_y;      // リフトの慣性(ジャンプ開始時にjump_yに代入)
                        // ※リフトの慣性は着地時にeb_push_leftなどで決まるが
                        //   ジャンプ中にはリフトから影響を受けないので、
                        //   ジャンプ中の慣性補正値とは区別する必要がある
    int jump_x;         // ジャンプ時の水平方向追加移動量(リフトのブースト分)
    int jump_y;         // ジャンプ時の上下方向追加移動量(リフトのブースト分)
    int float_time;     // ジャンプ中の滞空(上昇+水平)時間カウンタ
	int jump_y_limit;   // ジャンプでの到達位置上限
	int blast_x;
	// !!! 以下のフラグ類はビット化して1ワードに押し込むことを検討
    int is_landed;      // 着地していたら1
	int is_prev_landed;    // 前ターンに着地していたら1
    int is_walking;     // 前ターンに歩いていたら1
	int is_walk_blocked;// 前ターンに歩行が壁でブロックされていたら1
	int is_sliding;     // 滑り落ちているなら1
	// !!! stunnedは特定のstateと同時に発生するので、state側の特定ビットで表現する
	//    int is_stunned;     // 行動不能状態なら1
} obj_creature_t;

#define IsCreature(obj) ((obj)->type <= ot_chee)
#define MoveX(obj) ((obj)->u.creature.move_x)
#define SetMoveX(obj, val)  ((obj)->u.creature.move_x = (val))
#define SetMoveXDir(obj) ((obj)->u.creature.move_x = \
                          ((obj)->dir == di_left ? -1 : 1))
#define InertiaX(obj) ((obj)->u.creature.inertia_x)
#define SetInertiaX(obj, val)  ((obj)->u.creature.inertia_x = (val))
#define JumpX(obj) ((obj)->u.creature.jump_x)
#define SetJumpX(obj, val)  ((obj)->u.creature.jump_x = (val))
#define JumpY(obj) ((obj)->u.creature.jump_y)
#define SetJumpY(obj, val)  ((obj)->u.creature.jump_y = (val))
#define BlastX(obj) ((obj)->u.creature.blast_x)
#define SetBlastX(obj, val)  ((obj)->u.creature.blast_x = (val))
#define LbxOfst(obj) ((obj)->u.creature.land_ofst)
#define RbxOfst(obj) ((obj)->u.creature.land_ofst)
#define LandLx(obj) (Lbx(obj)+LbxOfst(obj))
#define LandRx(obj) (Rbx(obj)-RbxOfst(obj))
#define LandLpx(obj) (Lbpx(obj)+LbxOfst(obj))
#define LandRpx(obj) (Rbpx(obj)-RbxOfst(obj))
#define LandWidth(obj) ((obj)->width - (LbxOfst(obj)+RbxOfst(obj)))

//#define IsStunned(obj) ((obj)->u.creature.is_stunned)
//#define SetStunned(obj, val) ((obj)->u.creature.is_stunned = (val))
#define IsLanded(obj) ((obj)->u.creature.is_landed)
#define SetLanded(obj, val) ((obj)->u.creature.is_landed = (val))
#define IsPrevLanded(obj) ((obj)->u.creature.is_prev_landed)
#define SetPrevLanded(obj, val) ((obj)->u.creature.is_prev_landed = (val))
#define IsSliding(obj) ((obj)->u.creature.is_sliding)
#define SetSliding(obj, val) ((obj)->u.creature.is_sliding = (val))
#define IsWalkBlocked(obj) ((obj)->u.creature.is_walk_blocked)
#define SetWalkBlocked(obj, val) ((obj)->u.creature.is_walk_blocked = (val))
#define IsWalking(obj) ((obj)->u.creature.is_walking)
#define SetWalking(obj, val) ((obj)->u.creature.is_walking = (val))
#define IsRooted(obj) (CheckAttr(obj, at_rooted))
#define IsMeat(obj) (CheckAttr(obj, at_meat))
#define IsVegetable(obj) (CheckAttr(obj, at_vegetable))
#define IsImmuneFlame(obj) (CheckAttr(obj, at_immune_flame))
#define IsImmuneWater(obj) (CheckAttr(obj, at_immune_water))
#define CanClimb(obj)  (CheckAttr(obj, at_can_climb))
#define IsTurnToMist(obj) (CheckAttr(obj, at_turn_to_mist))


#define GetDefaultMotion(obj, member) (creature_image_table[(obj)->type].member)
#define GetMotion(obj) ((obj)->motion)
#define SetMotion(obj, m) ((obj)->motion = (m))
#define SetMotionDir(obj, ml, mr)             \
    ((obj)->dir == di_left ?                  \
     ((obj)->motion = (ml)) :                 \
     ((obj)->motion = (mr)))
#define SetWalkMotion(obj)                                         \
    (((obj)->dir == di_left) ?                                     \
     ((GetMotion(obj) == GetDefaultMotion(obj, walk_l0)) ?         \
      SetMotion((obj), GetDefaultMotion((obj), walk_l1)) :         \
      SetMotion((obj), GetDefaultMotion((obj), walk_l0))) :        \
     ((GetMotion(obj) == GetDefaultMotion(obj, walk_r0)) ?         \
      SetMotion((obj), GetDefaultMotion((obj), walk_r1)) :         \
      SetMotion((obj), GetDefaultMotion((obj), walk_r0))))


typedef struct {
	int no; // signメッセージの番号
} obj_sign_t;

typedef struct {
    int is_reverse; // 1なら踏んでいるときOFF
} obj_mswitch_t;

typedef struct {
	int lenctr;
} obj_flame_t;

struct obj_s;

typedef enum {lo_invalid, lo_and, lo_or, lo_xor, lo_majority} logic_t;
typedef enum {oo_off = -1, oo_invalid = 0, oo_on = 1} onoff_t;

typedef struct {
	logic_t logic;
    int is_reverse;
    int on_num;
	int in_num;
	int out_num;
	struct obj_s *in_objs[HUB_IN_MAX];
	struct obj_s *out_objs[HUB_OUT_MAX];
    unsigned int in_flags;
} obj_hub_t;

#define SetHubInFlag(obj, ix) ((obj)->u.hub.in_flags |= 1 << (ix))
#define ResetHubInFlag(obj, ix) ((obj)->u.hub.in_flags &= ~(1 << (ix)))
#define CheckHubInFlag(obj, ix) ((obj)->u.hub.in_flags & 1 << (ix))

/*
typedef struct {
    int beam_height;
} obj_teleporter_t;

typedef struct {
} obj_chee_t;

typedef struct {
} obj_hatopo_t;

typedef struct {
    int width;       // 水平方向の大きさ
    int xmin, xmax;  // 水平位置(obj->x)の最小・最大値
} obj_mvfloor_t;

*/

/*
typedef struct {
    int width;       // 水平方向の大きさ  
    int ymin, ymax;  // 垂直位置(obj->y)の最小・最大値
} obj_lift_t;
*/

#define STATE_STACK_SIZE 4

typedef struct obj_s{
    obj_class_t class;
    obj_type_t type;
    dir_t dir;
    int x, y;
	int prev_x, prev_y;
    int width;
    int height;
	struct obj_s *link;  // !! 将来的に配列化? 
    int motion;         // 動きパターンのインデックス
    obj_state_t state; // 死亡、食べている、などの状態
    obj_state_t prev_state; // １ターン前の状態
    obj_state_t state_stack[STATE_STACK_SIZE];
	int ctr_stack[STATE_STACK_SIZE];
	int state_stack_top;
    //    obj_state_t prev_state;
    int ctr;              // 複数ステップかかるstate用のカウンタ
    struct obj_s *prev, *next;
    int attr;
    union {
        obj_creature_t creature;
		obj_sign_t sign;
		obj_mswitch_t mswitch;
		obj_flame_t flame;
		obj_hub_t hub;
        //        obj_teleporter_t teleporter;
//		obj_nya_t nya;
        //        obj_mvfloor_t mvfloor;
        //        obj_lift_t lift;
    } u;
} obj_t;

typedef struct {
	// オブジェクト生成時にオブジェクト固有の初期化を追加する
	// ※set_emap_staticはここで行ってもよいが、
	// 　set_emapはすぐに上書きされるので効果がない
	// ※自身よりマップ上で右・下方向のオブジェクトは未生成なので、
	// 　近接オブジェクトを参照するような処理をここで行ってはならない。
    void (*create)(obj_t *, int, int, dir_t);
	// フレーム毎のオブジェクトの動作を行う
	// 死亡などの場合は非0を返すとそのオブジェクトが削除される
	int (*setup)(obj_t *);
	int (*act)(obj_t *);  
	// そのオブジェクトを描画する
	int (*draw)(obj_t *);
	// そのオブジェクトの追加描画。
	// 各オブジェクトのdrawが呼ばれた後で実行される
	// 各オブジェクトのdraw呼び出し順はオブジェクト生成順に依存するので、
	// 他のオブジェクトの上から描きたい場合はこちらに処理を記述する
    int (*post_draw)(obj_t *);
	// オブジェクト固有の動作と、歩行など一般的な挙動の設定を行う
	// actから呼び出されることを想定しており、現状はオブジェクトが
	// creatureの場合にact_creatureからよばれる。
    int (*set_action)(obj_t *);
    void (*set_motion)(obj_t *);
	void (*set_emap)(obj_t *);
} obj_funcs_t;

typedef struct {
    obj_type_t type;
    char *name;
    obj_class_t oclass;
    obj_funcs_t funcs;
} obj_info_t;

// 他のオブジェクトからのstate変更リクエストキュー
// ※直接stateを変更させると、要求側と変更側の処理順に依存して
// 状態遷移タイミングがずれるので、
// 一端キューにため込んで、ステップ毎に一括処理する
typedef struct {
    obj_t *dest;
    obj_t *src;
    obj_state_t state;
} state_req_entry_t;

#define State(obj) ((obj)->state)
#define PrevState(obj) ((obj)->prev_state)
#define SavePrevState(obj) ((obj)->prev_state = (obj)->state)

/*
#define PrevState(obj) ((obj)->prev_state)
#define SetState(obj, st)                    \
    (((obj)->prev_state = (obj)->state),     \
     ((obj)->state = st),                    \
     ((obj)->ctr = 0))
*/
#define Ctr(obj) ((obj)->ctr)
#define SetState(obj, st)                    \
    (((obj)->state_stack_top = 0),           \
     ((obj)->state = st),                    \
     ((obj)->ctr = 0))

#define PushState(obj, st)                                   \
    (assert((obj)->state_stack_top < STATE_STACK_SIZE),           \
     ((obj)->state_stack[(obj)->state_stack_top] = (obj)->state), \
     ((obj)->ctr_stack[(obj)->state_stack_top++] = (obj)->ctr), \
     ((obj)->state = st),                                    \
     ((obj)->ctr = 0))

#define PopState(obj)                                        \
    (assert((obj)->state_stack_top > 0),                            \
     ((obj)->state = (obj)->state_stack[--((obj)->state_stack_top)]), \
     ((obj)->ctr = (obj)->ctr_stack[(obj)->state_stack_top]))

#define PeekState(obj)                          \
    (assert((obj)->state_stack_top > 0),               \
     ((obj)->state_stack[(obj)->state_stack_top-1]))


#define SetCtr(obj, val)                        \
    ((obj)->ctr = val)
#define IncCtr(obj) ((obj)->ctr ++)

#define IsAlive(obj) (State(obj) & OS_ALIVE)
#define IsStunned(obj) (State(obj) & OS_STUN)
#define IsWalkable(obj) (State(obj) == os_none)
#define IsBlasted(obj) (State(obj) == os_blasted)


//    (State(obj) >= os_none &&                 
//     State(obj) < os_is_alive_end)

#define IsEating(obj) (State(obj) == os_eating)

#define ReqState(dest, src, st)                  \
    (push_obj_req(dest, src, st))




void obj_init(void);
void setup_obj_list(void);
obj_t *create_obj_grid(obj_type_t type, int x, int y, dir_t dir);
obj_t *create_obj(obj_type_t type, int x, int y, dir_t dir);
void delete_obj(obj_t *obj);
char *obj_type2name(obj_type_t type);
obj_type_t obj_name2type(char *name);
obj_t *create_sign_grid(obj_type_t type, int x, int y, int link_no);


void forall_obj(int(*f)(obj_t *), obj_class_t class);
//void act_obj_list(void);
int act_obj(obj_t *obj);
void setup_stage_obj_list(void);
int setup_stage_obj(obj_t *obj);
void draw_obj_list(void);
int draw_obj(obj_t *obj);
int post_draw_obj(obj_t *obj);

void clear_obj_req_queue(void);
void push_obj_req(obj_t *dest, obj_t *src, obj_state_t state);
void apply_req(obj_t *dest, obj_t *src, obj_state_t state);
void apply_all_req(void);

void create_chee(obj_t *obj, int ox, int oy, dir_t dir);
int setup_chee(obj_t *obj);
void set_chee(obj_t *obj);
obj_t *get_chee(void);
void create_gondin(obj_t *obj, int ox, int oy, dir_t dir);
int setup_gondin(obj_t *obj);
void create_hatopo(obj_t *obj, int ox, int oy, dir_t dir);
void create_yaramasi(obj_t *obj, int ox, int oy, dir_t dir);
void create_gagga(obj_t *obj, int ox, int oy, dir_t dir);
void create_ojigiri(obj_t *obj, int ox, int oy, dir_t dir);
void create_sue(obj_t *obj, int ox, int oy, dir_t dir);
void create_chu(obj_t *obj, int ox, int oy, dir_t dir);
void create_mingo(obj_t *obj, int ox, int oy, dir_t dir);
void create_jumeme(obj_t *obj, int ox, int oy, dir_t dir);
void create_siguku(obj_t *obj, int ox, int oy, dir_t dir);
void create_flame(obj_t *obj, int ox, int oy, dir_t dir);
void create_nya(obj_t *obj, int ox, int oy, dir_t dir);
void create_punch(obj_t *obj, int ox, int oy, dir_t dir);
void create_wall(obj_t *obj, int ox, int oy, dir_t dir);
void create_goal(obj_t *obj, int ox, int oy, dir_t dir);
void create_sign(obj_t *obj, int ox, int oy, dir_t dir);
void create_sting(obj_t *obj, int ox, int oy, dir_t dir);
void create_fwall(obj_t *obj, int ox, int oy, dir_t dir);
void create_iwall(obj_t *obj, int ox, int oy, dir_t dir);
void create_barrier(obj_t *obj, int ox, int oy, dir_t dir);
void create_pipe(obj_t *obj, int ox, int oy, dir_t dir);
void create_slope_l(obj_t *obj, int ox, int oy, dir_t dir);
void create_slope_r(obj_t *obj, int ox, int oy, dir_t dir);
void create_stairs_l(obj_t *obj, int ox, int oy, dir_t dir);
void create_stairs_r(obj_t *obj, int ox, int oy, dir_t dir);
void create_rstairs_l(obj_t *obj, int ox, int oy, dir_t dir);
void create_rstairs_r(obj_t *obj, int ox, int oy, dir_t dir);
//void create_mvfloor(obj_t *obj, int ox, int oy, dir_t dir);
//void set_grid_range_mvfloor(obj_t *obj, int xmin, int xmax);
//void set_grid_width_mvfloor(obj_t *obj, int width);
void create_lift(obj_t *obj, int ox, int oy, dir_t dir);
void create_mswitch(obj_t *obj, int ox, int oy, dir_t dir);
void create_aswitch(obj_t *obj, int ox, int oy, dir_t dir);
void create_teleporter(obj_t *obj, int ox, int oy, dir_t dir);
void create_teleport_dest(obj_t *obj, int ox, int oy, dir_t dir);
void create_hub(obj_t *obj, int ox, int oy, dir_t dir);
void set_hub_logic(obj_t *obj, logic_t logic);
void set_hub_reverse(obj_t *obj, int is_reverse);
void register_hub_in(obj_t *hub, obj_t *obj);
void register_hub_out(obj_t *hub, obj_t *obj);
void apply_req_to_hub(obj_t *hub, obj_t *src, obj_state_t state);
void print_hub(obj_t *hub);
void set_grid_width_static(obj_t *obj, int width);
void set_grid_width_dynamic(obj_t *obj, int width);
void set_grid_height_dynamic(obj_t *obj, int height);

//obj_state_t exclude_obj(obj_t *obj);
exclude_result_t exclude_obj(obj_t *obj);
int move_obj_horizontal(obj_t *obj, int xoffset);
int walk_obj(obj_t *obj, int xoffset);
int push_horizontal(obj_t *obj);
int push_diagonal(obj_t *obj);
int blast_horizontal(obj_t *obj);
int start_teleport_creature(obj_t *obj);
int teleport_creature(obj_t *obj);

int setup_creature(obj_t *obj);
int act_creature(obj_t *obj);
void set_emap_creature(obj_t *obj);
int act_chee(obj_t *obj);
void set_motion_chee(obj_t *obj);
void set_emap_chee(obj_t *obj);
int set_action_gondin(obj_t *obj);
void set_motion_gondin(obj_t *obj);
void set_emap_gondin(obj_t *obj);
int set_action_hatopo(obj_t *obj);
void set_motion_hatopo(obj_t *obj);
int set_action_yaramasi(obj_t *obj);
void set_motion_yaramasi(obj_t *obj);
void set_emap_yaramasi(obj_t *obj);
int set_action_ojigiri(obj_t *obj);
void set_motion_ojigiri(obj_t *obj);
int set_action_gagga(obj_t *obj);
void set_motion_gagga(obj_t *obj);
void set_emap_gagga(obj_t *obj);
int set_action_nya(obj_t *obj);
void set_motion_nya(obj_t *obj);
int set_action_sue(obj_t *obj);
void set_motion_sue(obj_t *obj);
int set_action_chu(obj_t *obj);
void set_motion_chu(obj_t *obj);
int set_action_mingo(obj_t *obj);
void set_motion_mingo(obj_t *obj);
void set_emap_mingo(obj_t *obj);
int set_action_jumeme(obj_t *obj);
void set_motion_jumeme(obj_t *obj);
void set_emap_jumeme(obj_t *obj);
int set_action_siguku(obj_t *obj);
void set_motion_siguku(obj_t *obj);
void set_emap_siguku(obj_t *obj);
int act_nya(obj_t *obj);
int setup_nya(obj_t *obj);
int act_punch(obj_t *obj);
void set_emap_punch(obj_t *obj);
//void move_mvfloor(obj_t *obj);
int act_lift(obj_t *obj);
int setup_lift(obj_t *obj);
void set_emap_lift(obj_t *obj);
int act_goal(obj_t *obj);
int act_fwall(obj_t *obj);
int act_iwall(obj_t *obj);
int act_pipe(obj_t *obj);
int act_barrier(obj_t *obj);
int setup_barrier(obj_t *obj);
void set_motion_barrier(obj_t *obj);
void set_emap_barrier(obj_t *obj);
int act_mswitch(obj_t *obj);
int setup_mswitch(obj_t *obj);
void on_aswitch(obj_t *obj);
int act_aswitch(obj_t *obj);
int setup_aswitch(obj_t *obj);
void set_emap_aswitch(obj_t *obj);
int act_teleporter(obj_t *obj);
int setup_teleporter(obj_t *obj);
void set_motion_teleporter(obj_t *obj);
void set_emap_teleporter(obj_t *obj);
int act_flame(obj_t *obj);
void set_emap_flame(obj_t *obj);

#endif
