#ifndef RECORD_H
#define RECORD_H
#include "config.h"
#include "font.h"
#include "obj.h"

typedef struct {
    unsigned int min;
    unsigned int max;
} minmax_t;

// 累積カウント型の記録エントリ
typedef enum {
    re_invalid = -1,
    re_begin = 0,
    re_score = 0,    // とくてん
    re_time,         // のこりじかん
    re_gondin_eaten, // ごんでぃんを食べた数
	re_gondin_lost,  // ごんでぃんを失った数
    re_clear,        // クリア回数。死亡回数はkilled_by_xxxから計算できる
    re_walk_frame,   // 歩行したフレーム数
    re_air_frame,    // 空中にいたフレーム数
    re_jump_count,   // ジャンプした回数
    re_slide_frame,   // すべったしたフレーム数
	re_lift_frame,   // リフト/にゃあぱんち上にいたフレーム数
	re_blast_count,  // がっがにふっとばされた回数
	re_switch_count,  // スイッチを切り替えた数
	re_teleport_count,  // テレポート回数

    // 死因別の死亡回数
    // 判定上、たとえばやらましと棘のどちらに刺さったかなどは区別されない
    re_killed_by_pierced,  // やらまし／棘にささった
    re_killed_by_ojigiri,  // おじぎり草に食べられた
    re_killed_by_smash,    // リフトに挟まれた
    re_killed_by_flame,    // 炎に焼かれた
	re_killed_by_barrier,  // バリアで感電した
	// 各キャラの死亡数
    re_hatopo_killed,
    re_yaramasi_killed,
    re_gagga_killed,
    re_ojigiri_killed,
    re_sue_killed,
    re_chu_killed,
    re_mingo_killed,
    re_jumeme_killed,
    re_siguku_killed,
    re_fwall_destroyed,
    re_iwall_destroyed,
    re_end = 256
} record_entry_t;

typedef struct {
    unsigned int min[re_end];
    unsigned int max[re_end];
    unsigned int sum[re_end];
} record_t;

typedef struct {
    record_entry_t entry;
    char *fontname;
    char *description;
    int show_minmax;
    int show_sum;
} record_info_t;

typedef struct {
	obj_type_t type;
	record_entry_t entry;
} ot2re_info_t;

#define RecordMin(r, re) (r.min[re])
#define RecordMax(r, re) (r.max[re])
#define RecordSum(r, re) (r.sum[re])

#define MINMAX_MAX 0x1000000

void record_init(void);
font_id_t *get_record_icon(record_entry_t re);
int get_record_showentry_num(void);
record_info_t *get_record_info(int no);
record_entry_t kill_ot2re(obj_type_t type);
char *get_record_description(record_entry_t entry);

#endif

