#include <stdio.h>
#include <EzGraph.h>
#include "config.h"
#include "main.h"
#include "layout.h"
#include "select_stage.h"
#include "select_stage_draw.h"
#include "stage.h"

char *stage_names[] = {
	"walk_and_jump",
	"jumpjump",
    "lift",
    "shortcut",
	"yaramasi",
    "sting",
    "hatopo",
    "ojigiri",
	"switch",
	"largejump",
	"gagga",
	"freefall",
	"slope",
	"tunnel",
    "lift_gondin",
	"nya",
	"barrier",
	"sue",
	"fragile",
	"ojigiriwars",
	"chu",
	"flags",
	"gondin",
	"ojigiriwars2",
	"knock",
	"pita",
	"nya_sting",
	"teleporter",
	"underground",
	"underground2",
	"longjump",
	"underground3",
	"jumeme",
	"heiankyo",
	"level6",
	"wiz",
	"siguku",
	"pita2",
	"hub",
    "title",
	"full_adder",
	"4bit_adder",
	"mingo",
	"birthday",
	"pyramid",
	"pyramid2",
	"raiders",
	"snail",
	"water_slider",
	"hatopo_xpm",
	"obstacle",
	"castle",
	"island_campaign",
	"chknpp",
	"pisa",
	"pisa2",
	"amida",
	"chase_gondin",
    "chu_sue",
    "siguku_adv",


#ifdef DEBUG
	"test2",

	"save_gondin",

	"ice",

	"rain",
	"fire",

	"test4",
	"chobbit",

	"gravity",
	"test",
    "test3",
	"slide",
	"ilikecats",
	"test5",
	"test6",
	"test7",
	"test8",
//	"heiankyo2",
	"snowman",
	"maxwell",
	"cake",
	"enquor",
	"run",
	"dig_tunnel",
	"test00",
	"test01",
	"test02",
	"test03",
	"test04",
	"test09",
/*
	*/
#endif
 };

#define STAGE_NUM (sizeof(stage_names)/sizeof(char*))

int total_score = 0;
int cleared_stage_num = 0;
int stage_num = STAGE_NUM;
int achieved_star_num = 0;
int total_star_num;
int line_num;

stage_thumb_t stage_thumbs[STAGE_NUM];
stage_thumb_t *current_stage_thumb = NULL;

static int cursor_x, cursor_y, line;


void select_stage_init(void){
    int i;

    for (i = 0 ; i < stage_num ; i++){
        //        printf("gen thumb:%s\n", get_stage_name(i));
        get_thumb(get_stage_name(i), &stage_thumbs[i], i+1);
        //        total_score += stage_thumbs[i].top_score;
        //        if (stage_thumbs[i].top_score > 0) cleared_stage_num ++;
    }
    line = 0;
	cursor_x = 0;
	cursor_y = 0;
    line_num = (stage_num-1) / THUMB_COL_NUM + 1;
    printf("line_num = %d\n", line_num);
    update_thumbs_info();
}

void setup_select_stage(void){
    EzSetKeyHandler(select_stage_keyh);
    EzSetKeyReleaseHandler(select_stage_key_releaseh);
    EzSetTimerHandler(select_stage_timerh, GAME_SPEED);
    update_thumbs_info();
}

void select_stage_keyh(int key){
    //    printf("line=%d/%d, (%d,%d)", line, line_num, cursor_x, cursor_y);

    switch (key){
    case EZ_Left:
        cursor_x = (cursor_x + THUMB_COL_NUM - 1) % THUMB_COL_NUM;
        break;
    case EZ_Right:
        cursor_x = (cursor_x + 1) % THUMB_COL_NUM;
        break;
    case EZ_Up:
        if (cursor_y == 0){
            if (line > 0){
                line --;
            }
            else {
                line = line_num-THUMB_ROW_NUM;
                cursor_y = THUMB_ROW_NUM-1;
            }
        }
        else {
            cursor_y --;
        }
        break;
    case EZ_Down:
        if (cursor_y == THUMB_ROW_NUM-1){
            if (line < line_num-THUMB_ROW_NUM){
                line ++;
            }
            else {
                line = 0;
                cursor_y = 0;
            }
        }
        else {
            cursor_y ++;
        }
        break;
    default:
        break;
    }
    printf(" -> %d/%d, (%d,%d)\n", line, line_num, cursor_x, cursor_y);
}

void select_stage_key_releaseh(int key){
    int si;
    switch (key){
    case ' ':
    case EZ_Return:
        si = CursorPos2StageIx(cursor_x, cursor_y, line);
#ifdef LOCK_STAGE
        if (si < stage_num && stage_thumbs[si].unlocked){
#else
        if (si < stage_num){
#endif
			current_stage_thumb = &stage_thumbs[si];
            load_current_stage(get_stage_name(si), si+1);
            set_main_state(ms_game);
        }
        break;
    case 'r':
        si = CursorPos2StageIx(cursor_x, cursor_y, line);
        if (si < stage_num){
			current_stage_thumb = &stage_thumbs[si];
            //            load_current_stage(get_stage_name(sn), sn+1);
            load_current_stage_record(get_stage_name(si), 
                                      current_stage_thumb->major_version, 
                                      current_stage_thumb->minor_version);
            set_main_state(ms_stage_info);
        }
        break;
    case 'q':
	case 27:
        set_main_state(ms_title);
		break;
    default:
        break;
    }
}

void select_stage_timerh(void){
    draw_select_stage(stage_thumbs, stage_num, cursor_x, cursor_y, line);
}

char *get_stage_name(int n){
    return stage_names[n];
}


unsigned int get_current_stage_no(void){
    return CursorPos2StageIx(cursor_x, cursor_y, line) + 1;
}

stage_thumb_t *get_current_thumb(void){
	return current_stage_thumb;
}

unsigned int get_total_score(void){
    return total_score;
}
unsigned int get_total_star_num(void){
    return total_star_num;
}
unsigned int get_achieved_star_num(void){
    return achieved_star_num;
}
unsigned int get_cleared_stage_num(void){
    return cleared_stage_num;
}
unsigned int get_stage_num(void){
    return stage_num;
}
// サムネイル一覧表示用の情報を、最新のものに更新する
void update_thumbs_info(void){
    int i;
#ifdef LOCK_STAGE
    int unlock_num;
#endif

    total_score = 0;
    cleared_stage_num = 0;
    achieved_star_num = 0;
    total_star_num = 0;
    for (i = 0 ; i < stage_num ; i++){
        total_score += stage_thumbs[i].top_score;
        total_star_num += get_thumb_star_num(&stage_thumbs[i]);
        if (stage_thumbs[i].top_score > 0){
            cleared_stage_num ++;
            achieved_star_num += get_thumb_achieved_star_num(&stage_thumbs[i]);
        }
#ifdef LOCK_STAGE
        stage_thumbs[i].unlocked = (stage_thumbs[i].clear_num > 0) ? 1 : 0;
#endif
    }
#ifdef LOCK_STAGE
    // 挑戦可能なステージ数はクリア数0-4で1、クリア数5-9で2、...
    unlock_num = cleared_stage_num / 5 + 1;
    printf("cleared_stage_num:%d unlock_num:%d\n", cleared_stage_num, unlock_num);
    for (i = 0 ; i < stage_num ; i++){        
        if (stage_thumbs[i].unlocked == 0){
            stage_thumbs[i].unlocked = 1;
            unlock_num --;
            if (unlock_num == 0) break;
            printf("unlock stage %d unlock:%d\n", i+1, unlock_num);
        }
    }
#endif
}

#ifdef LOCK_STAGE
int is_obj_introduced(obj_type_t type){
    int stage;

    stage = get_introduction_stage(type);
    return stage_thumbs[stage-1].unlocked;
}
#endif
