#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <EzGraph.h>
#include "config.h"
#include "main.h"
#include "game.h"
#include "stage.h"
#include "select_stage.h"
#include "emap.h"
#include "obj.h"
#include "obj_draw.h"
#include "file.h"
#include "jstr.h"

#define BUF_SIZE 1024

static stage_t current_stage;
static char stage_dpath[BUF_SIZE];
static record_t current_stage_record;
static unsigned int obj_introduction[ot_end];

typedef struct {
    unsigned char symbol;
    int has_option; // オプションをとるものは非0
    obj_type_t type;
} map_symbol_info_t;

static map_symbol_info_t map_symbol_info_list[] = {
    {' ', 0, ot_blank},
    {'.', 0, ot_blank},
    {'~', 1, ot_teleport_dest},
    {'#', 0, ot_wall},
    {'%', 0, ot_fwall},
    {'*', 0, ot_iwall},
    {'$', 1, ot_barrier},
    {'/', 0, ot_slope_l},
    {'\\', 0, ot_slope_r},
    {'d', 0, ot_stairs_l},
    {'b', 0, ot_stairs_r},
    {'q', 0, ot_rstairs_l},
    {'p', 0, ot_rstairs_r},
    {'=', 1, ot_lift},
    {'-', 1, ot_pipe},

	{'S', 1, ot_sign},
	{'s', 1, ot_sign_hang},
    {'|', 1, ot_sting},
    {'D', 0, ot_goal},
    {'?', 1, ot_mswitch},
    {'!', 1, ot_aswitch},
    {'&', 1, ot_teleporter},
    {'"', 1, ot_hub},

	{'C', 1, ot_chee},
    {'G', 0, ot_gondin},
    {'H', 1, ot_hatopo},
    {'K', 1, ot_gagga},
    {'M', 1, ot_chu},
    {'I', 1, ot_mingo},
    {'N', 1, ot_nya},
    {'O', 1, ot_ojigiri},
    {'U', 1, ot_sue},
    {'Y', 1, ot_yaramasi},
    {'F', 1, ot_jumeme},
};
#define MAP_SYMBOL_INFO_NUM \
    (sizeof(map_symbol_info_list)/sizeof(map_symbol_info_t))

static obj_type_t symbol2type_table[128];
static int type2has_option_table[ot_end];
#define HasOption(type) (type2has_option_table[type])

static char link_chrs[] = LINK_CHRS_STR;

#define LinkNo2Chr(no) (link_chrs[no])

////static link_entry_t link_entries[LINK_NO_MAX];
//// リンクは多対一可なため、src側はリンク番号とペアですべて保存し、
//// dest側はリンク番号をキーにしたテーブルに登録する
// ハブによる多対多接続を可能にするため、
// src, destともいったんリンク番号とペアですべて保存する。
static link_entry_t link_src_entries[LINK_NUM_MAX];
static int link_src_entry_num;
static link_entry_t link_dest_entries[LINK_NUM_MAX];
static int link_dest_entry_num;

typedef struct {
    dir_t dir;
    onoff_t onoff;
    // テレポーター、ハブなど2リンク必要なものもあるが、
    // 入出力リンクの順序はオブジェクト依存なので、
    // ここでは見つかった順にno1, no2として保持
    int link_no1; 
    int link_no2;
	logic_t logic;
} map_options_t;


static objective_t stage_objectives[OBJECTIVE_NUM];

void clear_stage(stage_t *stage);
void get_obj_options(stage_t *stage, int x, int y, obj_type_t target_type,
                     map_options_t *options);

dir_t get_map_objdir(stage_t *stage, int x, int y);

void stage_init(char *argv0){
    int i;
    obj_type_t type;
    for (i = 0 ; i < MAP_SYMBOL_INFO_NUM ; i++){
        type = map_symbol_info_list[i].type;
        symbol2type_table[map_symbol_info_list[i].symbol] = type;
        type2has_option_table[type] = map_symbol_info_list[i].has_option;
        //        printf("%c[%d] -> %d(%d)\n", map_symbol_info_list[i].symbol,
        //               map_symbol_info_list[i].symbol, map_symbol_info_list[i].type,
        //               symbol2type_table[map_symbol_info_list[i].symbol]);
    }

	// Visual Studioでの開発中は、
	// 実行ファイルがソースと異なるディレクトリに作られるので、
	// 実行ファイルからの相対ディレクトリ指定では正しく動作しない
	// カレントディレクトリはソース(プロジェクトファイル?)の場所なので、
	// とりあえずカレントからの相対指定にしておく
#if defined(_MSC_VER) && defined(_DEBUG)
    sprintf(stage_dpath, "%s", STAGE_DIR);
#else
    sprintf(stage_dpath, "%s/%s", get_exe_dpath(argv0), STAGE_DIR);
    printf("dpath:%s\n", stage_dpath);
#endif

    for (type = ot_begin ; type < ot_end ; type++){
        obj_introduction[type] = -1;
    }
}


void load_current_stage(char *name, int stage_no){
    //    printf("load current stage %s\n", stage_dpath);
    load_stage(name, &current_stage, stage_no);
	current_stage.no = stage_no;
    load_current_stage_record(name, 
                              current_stage.major_version,
                              current_stage.minor_version);
}


void clear_stage(stage_t *stage){
	int i, j;

    for (i = 0 ; i < STAGE_TITLE_ROW_NUM ; i++){
        stage->title[i][0] = fi_end;
    }
	stage->gondin_toeat = 1;
    for (i = 0 ; i < OBJECTIVE_NUM ; i++){
        stage_objectives[i].entry = re_invalid;
    }
    for (i = 0 ; i < LINK_NO_MAX ; i++){
        for (j = 0 ; j < STAGE_SIGN_ROW_NUM ; j++){
            current_stage.signs[i][j][0] = fi_end;
        }
    }
}

int load_stage(char *stage_name, stage_t *stage, int stage_no){
    FILE *fp;
    char fpath[BUF_SIZE];
    char tline[BUF_SIZE];
    stage_command_t sc = sc_none;
    char *scp;
    char link_chr;
    int link_no = 0;
    int line_no = 0, map_row = 0, title_row = 0, sign_row = 0;
    static char *sc_str = SC_STR;
    char name[BUF_SIZE];
    char opchr;
    unsigned int val;
    int objective_num = 0;
	int id;
    objective_info_t *objective_info;
	unsigned int v1, v2;

	clear_stage(stage);
	stage->name = stage_name;
    sprintf(fpath, "%s/%s.%s", stage_dpath,  stage_name, STAGE_EXT);
    printf("read:%s...", fpath);

    fp = fopen_bomcut(fpath, "r");
    if (fp == NULL){
        log_error("Cannot open stage file %s.\n", fpath);
        exit(1);
    }
    while (fgets(tline, BUF_SIZE, fp) != NULL){
        //        fprintf(stdout, "%d:%s", line_no, tline);
        line_no ++;
        if (tline[0] == '@'){
            scp = strchr(sc_str, tline[1]);
            if (scp == NULL){
                log_error("Unknown command @%c at line %d in %s\n",
                       tline[1], line_no, stage_name);
                exit(1);
            }
            sc = scp - sc_str+1;
            //            printf("stage command %d\n", sc);
            if (sc == sc_sign){
                if (sscanf(tline, "@sign %c", &link_chr) != 1){
                    log_error("Illegal sign line at line %d in %s\n",
                              line_no, stage_name);
                    exit(1);
                }
                link_no = chr2link_no(link_chr);
                sign_row = 0;
                //                printf("sign chr:%d(%c)\n", link_no, link_chr);
            }
			else if (sc == sc_gondin){
				if (sscanf(tline, "@gondin %d", &stage->gondin_toeat) != 1){
                    log_error("Illegal gondin line at line %d in %s\n",
                              line_no, stage_name);
                    exit(1);
                }
			}
			else if (sc == sc_version){
				if (sscanf(tline, "@version %d.%d", &v1, &v2) != 2){
                    log_error("Illegal version line at line %d in %s\n",
                              line_no, stage_name);
                    exit(1);
                }
				if (v1 < 256 && v2 < 256){
					stage->major_version = v1;
					stage->minor_version = v2;
				}
				else {
                    log_error("Illegal version value %d.%d at line %d in %s\n",
                              v1, v2, line_no, stage_name);
                    exit(1);
				}
			}
        }
        else {
            // 改行コードを削除
            chop_line(tline);
            // 行頭#か空行は読み飛ばす
			if (sc != sc_map && tline[0] == '#') continue;
            if (*tline == '\0') continue;

            switch (sc){
            case sc_title:
                if (title_row >= STAGE_TITLE_ROW_NUM){
                    log_error("Too many title lines at line %d in %s\n",
                              line_no, stage_name);
                    exit(1);
                }
                str2fontseq(tline, stage->title[title_row]);
                title_row++;
                break;
            case sc_description:
                break;
            case sc_map:
                if (map_row >= MAP_GRID_HEIGHT){
                    log_error("Too many lines in map at line %d in %s\n",
                              line_no, stage_name);
                    exit(1);
                }
                // マップ行の長さがMAP_GRID_WIDTH*2+1でない場合のチェックが必要
//                printf("strncpy %d %d \n", strlen(tline), MAP_GRID_WIDTH*2+1);
                strncpy(stage->map[map_row], tline, MAP_GRID_WIDTH*2+1);
                //                                strcpy(stage->map[map_row], tline);
                map_row++;
                break;
            case sc_sign:
                if (sign_row == STAGE_SIGN_ROW_NUM){
                    log_error("Too many lines (limit: %d) in sign %d "
                              "at line %d in %s\n",
                              STAGE_SIGN_ROW_NUM, link_no, line_no, stage_name);
                    exit(1);
                }
                //                printf("link_no=%d sign_row=%d\n", link_no, sign_row);
                str2fontseq(tline, stage->signs[link_no][sign_row]);
                sign_row ++;
				break;
            case sc_introduction:
                { 
                    obj_type_t type;
                    type = obj_name2type(tline);
                    //                    printf("tline:%s, type:%d\n", tline, type);
                    if (type == ot_none){
                        log_error("Unknown object '%s'"
                                  " at line %d in stage '%s'.\n",
                                  tline, line_no, stage_name);
                        exit(1);
                    }
                    obj_introduction[type] = stage_no;
                }
                break;
            case sc_objective:
                if (sscanf(tline, "%s %c= %d", name, &opchr, &val) != 3){
                    log_error("Illegal objective line",
                              " at line %d in stage '%s'.\n",
                              line_no, stage_name);
                    exit(1);
                }
                //                printf("mission:%s:%d %c %d\n",
                //                       tline, objective_name2id(name), opchr, val);
                if (objective_num >= OBJECTIVE_NUM){
                    log_error("Too many objectives"
                              " at line %d in stage '%s'\n",
                              line_no, stage_name);
                    exit(1);
                }
				id = objective_name2id(name);
				if (id < 0){
					log_error("Unknown objective '%s'"
                              " at line %d in stage '%s'\n",
                              name, line_no, stage_name);
                    exit(1);
				}
                objective_info = get_objective_info(id);
				stage_objectives[objective_num].entry
                    = objective_info->entry;
				stage_objectives[objective_num].icon
                    = objective_info->icon;
                stage_objectives[objective_num].incdec
                    = objective_info->incdec;
                stage_objectives[objective_num].description_fontseq
                    = objective_info->description_fontseq;
                switch (opchr){
                case '>':
                    stage_objectives[objective_num].op = oo_ge;
                    break;
                case '<':
                    stage_objectives[objective_num].op = oo_le;
                    break;
                case '=':
                    stage_objectives[objective_num].op = oo_eq;
                    break;
                default:
                    log_error("Unknown objective operator %c "
                              " at line %d in stage '%s'\n",
                              opchr, line_no, stage_name);
                    exit(1);
                    break;
                }
                stage_objectives[objective_num].val = val;
                objective_num ++;
                break;
            default:
                assert(0);
                exit(1);
            }
        }
    }
    /*
    for (i = 0 ; i < MAP_GRID_HEIGHT ; i++){
        fgets(stage->map[i], MAP_GRID_WIDTH*2+2, fp);
        fprintf(stdout, "%d:%s", i, stage->map[i]);
    }
    */
    fclose(fp);
    return -1;
}

font_id_t *get_current_stage_title(int row){
    return current_stage.title[row];
}
char *get_current_stage_name(void){
    return current_stage.name;
}
int get_current_stage_sign_no(void){
	return current_stage.sign_no;
}
void set_current_stage_sign_no(int no){
	current_stage.sign_no = no;
}

font_id_t *get_current_stage_sign(int row){
	assert(current_stage.sign_no != INVALID_LINK_NO);
    return current_stage.signs[current_stage.sign_no][row];
}
int get_current_stage_gondin_toeat(void){
	return current_stage.gondin_toeat;
}
int get_current_stage_gondin_total(void){
	return current_stage.gondin_total;
}

//unsigned int get_current_stage_no(void){
//	return current_stage.no;
//}

obj_type_t symbol2type(unsigned char symbol){
    //    printf("<<%d:%d>>", symbol, symbol2type_table[symbol]);
    return symbol2type_table[symbol];
}

int get_thumb_star_num(stage_thumb_t *thumb){
    int i, num = 0;
	for (i = 0 ; i < OBJECTIVE_NUM ; i++){
        if (thumb->stars_fontseq[i] != fi_space) num ++;
    }
    return num;
}
int get_thumb_achieved_star_num(stage_thumb_t *thumb){
    int i, num = 0;
	for (i = 0 ; i < OBJECTIVE_NUM ; i++){
        if (thumb->stars_fontseq[i] == fi_star_on) num ++;
    }
    return num;
}

void set_thumb_stars(stage_thumb_t *thumb){
	int i;
	objective_t *objective;

	for (i = 0 ; i < OBJECTIVE_NUM ; i++){
		objective = get_stage_objective(i);
		if (objective->entry == re_invalid){
			thumb->stars_fontseq[i] = fi_space;
		}
		else if (objective->op == oo_ge){
			if (CurrentStageRecordMax(objective->entry) >= objective->val){
				thumb->stars_fontseq[i] = fi_star_on;
			}
			else {
				thumb->stars_fontseq[i] = fi_star_off;
			}
		}
		else {
			if (CurrentStageRecordMin(objective->entry) <= objective->val){
				thumb->stars_fontseq[i] = fi_star_on;
			}
			else {
				thumb->stars_fontseq[i] = fi_star_off;
			}
		}
	}
	thumb->stars_fontseq[OBJECTIVE_NUM] = fi_end;
}

void get_thumb(char *stage_name, stage_thumb_t *thumb, int stage_no){
    int x, y, i;
    obj_type_t type;
    
    load_current_stage(stage_name, stage_no);
    for (i = 0 ; i < STAGE_TITLE_ROW_NUM ; i++){
        fontseq_cpy(thumb->title[i], current_stage.title[i]);
    }
	thumb->top_score = CurrentStageRecordMax(re_score);
	thumb->clear_num = CurrentStageRecordMax(re_clear);
	set_thumb_stars(thumb);
    thumb->major_version = current_stage.major_version;
    thumb->minor_version = current_stage.minor_version;
	
	
	
	for (y = 0 ; y < MAP_GRID_HEIGHT ; y++){
        for (x = 0 ; x < MAP_GRID_WIDTH ; x++){
            type = symbol2type(GetMapSymbol(x, y));
            switch(type){
            case ot_wall:
            case ot_fwall:
            case ot_stairs_l:
            case ot_stairs_r:
            case ot_rstairs_l:
            case ot_rstairs_r:
            case ot_slope_l:
            case ot_slope_r:
                SetThumbColor(x, y, 2);
				break;
            case ot_iwall:
                SetThumbColor(x, y, 7);
				break;
            case ot_pipe:
                SetThumbColor(x, y, 1);
				break;
            case ot_barrier:
                SetThumbColor(x, y, 5);
				break;
            case ot_lift:
            case ot_mswitch:
            case ot_aswitch:
            case ot_teleporter:
                SetThumbColor(x, y, 4);
				break;
            case ot_sting:
                SetThumbColor(x, y, 3);
				break;
            case ot_chee:
				set_thumb_color_box(x, y, 2, 2, 5, thumb);
				break;
            case ot_goal:
				set_thumb_color_box(x, y, 2, 2, 5, thumb);
				break;
            case ot_gondin:
				set_thumb_color_box(x, y, 2, 2, 6, thumb);
				break;
            case ot_nya:
				set_thumb_color_box(x, y, 2, 4, 1, thumb);
				break;
            case ot_ojigiri:
				set_thumb_color_box(x, y, 2, 4, 1, thumb);
				break;
            case ot_sue:
				set_thumb_color_box(x, y, 3, 2, 1, thumb);
				break;
            case ot_chu:
				set_thumb_color_box(x, y, 1, 1, 1, thumb);
				break;
            case ot_yaramasi:
				set_thumb_color_box(x, y, 2, 1, 1, thumb);
				break;
            case ot_hatopo:
            case ot_gagga:
            case ot_mingo:
            case ot_jumeme:
				set_thumb_color_box(x, y, 2, 2, 1, thumb);
				break;
            case ot_teleport_dest:
            case ot_hub:
            default:
                break;
            }
        }
    }
}

void set_thumb_color_box(int ltx, int lty, int width, int height, int color, stage_thumb_t *thumb){
	int x, y;
	for (y = lty ; y < lty+height ; y++){
		for (x = ltx ; x < ltx+width ; x++){
			SetThumbColor(x, y, color);
		}
	}
}

void get_obj_options(stage_t *stage, int x, int y, obj_type_t target_type,
                map_options_t *options)
{
    obj_type_t type = target_type; // (x, y)はtarget_typeのはず
    char c;
    int no;

    options->dir = di_none;
    options->onoff = oo_invalid;
    options->link_no1 = INVALID_LINK_NO;
    options->link_no2 = INVALID_LINK_NO;
	options->logic = lo_invalid;

    // このオブジェクトが連続するか、空白・壁などオプションなしオブジェクトであれば
    // オプションを列挙し続けられる
    // 異なるオブジェクトに切り替わったらそこで終了
    // 異なるオブジェクトが並ぶ場合も、8x8のオブジェクト以外なら
    // map上は空白が空くからそこでとぎれる。
    // chu, pipeなどは複数の同一オブジェクトがmap上で連続する可能性があるので
    // while内で同じオプションが出現したらエラーや上書きにせず、
    // 最初に見つけた同種オプションを有効とする
    // !! 「chu, pipe等の場合はループ2週目以降はtype==target_typeの判定を
    //    なしにする("M<M", "-^-^"などと書いたら、x+1の時点でループを抜ける)
    //    方がいいかもしれない。
    while (type == target_type || !HasOption(type)){
        c = GetMapOption(x, y);
        if (c == ' ') return; //オプションの列挙がとぎれたら終了
        switch (c){
        case '<':
            if (options->dir != di_none) return;
            options->dir = di_left;
            break;
        case '>':
            if (options->dir != di_none) return;
            options->dir = di_right;
            break;
        case '^':
            if (options->dir != di_none) return;
            options->dir = di_up;
            break;
        case 'v':
            if (options->dir != di_none) return;
            options->dir = di_down;
            break;
        case '+':
            if (options->onoff != oo_invalid) return;
            options->onoff = oo_on;
            break;
        case '-':
            if (options->onoff != oo_invalid) return;
            options->onoff = oo_off;
            break;
        case '&':
            if (options->logic != lo_invalid) return;
            options->logic = lo_and;
            break;
        case '|':
            if (options->logic != lo_invalid) return;
            options->logic = lo_or;
            break;
        case '%':
            if (options->logic != lo_invalid) return;
            options->logic = lo_xor;
            break;
        case '?':
            if (options->logic != lo_invalid) return;
            options->logic = lo_majority;
            break;
        default:
            no = chr2link_no(c);
            if (no == INVALID_LINK_NO){
                log_error("Unknown option '%c' at (%d, %d) in stage '%s'.\n",
                          c, x, y, current_stage.name);
                exit(1);
            }
            if (options->link_no1 == INVALID_LINK_NO){
                options->link_no1 = no;
            }
            else if (options->link_no2 == INVALID_LINK_NO){
                options->link_no2 = no;
            }
            else { // 3つめのリンク指定が見つかった
                return ;
            }
            break;
        }
        x++;
        if (x == MAP_GRID_WIDTH) break; // マップのその行の右端まで来たら終了
        type = symbol2type(GetMapSymbol(x, y));
    }
}



dir_t get_map_objdir(stage_t *stage, int x, int y){
    switch (GetMapOption(x, y)){
    case '<':
        return di_left;
    case '>':
        return di_right;
    case '^':
        return di_up;
    case 'v':
        return di_down;
    case ' ':
        return di_none;
    default:
        break;
        //        log_error("Unknown map option '%c' at (%d, %d).\n",
        //                  GetMapOption(x, y), x, y);
    //        exit(1);
    }
	return di_none;
}

int chr2link_no(char c){
    char *p;
    p = strchr(link_chrs, c);
    if (p == NULL) return INVALID_LINK_NO;
    return p - link_chrs;
}

int get_map_link(stage_t *stage, int x, int y){
    char c;
    int no;

	if (x >= MAP_GRID_WIDTH) return INVALID_LINK_NO;
    c = GetMapOption(x, y);
    no = chr2link_no(c);
	if (no == INVALID_LINK_NO){
		if (x+1 >= MAP_GRID_WIDTH) return INVALID_LINK_NO;
        c = GetMapOption(x+1, y);
        no = chr2link_no(c);
	}
    printf("got link no=%d (%c)\n", no, c);
    return no;
}

void register_link_src(obj_t *src, int link_no){
    assert(link_no != INVALID_LINK_NO);
    //    log_warn("register link src %d(%c) %d(%p)\n", link_no, LinkNo2Chr(link_no), src->type, src);
    link_src_entries[link_src_entry_num].obj = src;
    link_src_entries[link_src_entry_num].link_no = link_no;
    link_src_entry_num ++;
}
void register_link_dest(obj_t *dest, int link_no){
    assert(link_no != INVALID_LINK_NO);
    //    log_warn("register link dest %d(%c) %d(%p)\n", link_no, LinkNo2Chr(link_no), dest->type, dest);
    link_dest_entries[link_dest_entry_num].obj = dest;
    link_dest_entries[link_dest_entry_num].link_no = link_no;
    link_dest_entry_num ++;
}
/*
void register_link_dest(obj_t *dest, int link_no){
    assert(link_no != INVALID_LINK_NO);
    printf("register link %d dest %p\n", link_no, dest);
    link_dest_entries[link_no].dest = dest;
}
*/

typedef struct {
    int src_num;
    int dest_num;
    obj_t *srcs[MULTI_LINK_MAX];
    obj_t *dests[MULTI_LINK_MAX];
} link_table_entry_t;

void apply_init_req(obj_t *dest, obj_t *src){
    switch (src->type){
    case ot_aswitch:
        printf("apply aswitch\n");
        if (State(src) == os_nonweighted_on){
            apply_req(dest, src, os_on);
        }
        else {
            apply_req(dest, src, os_off);
        }
        break;
    case ot_mswitch:
        printf("apply mswitch\n");
        if (src->u.mswitch.is_reverse){
            apply_req(dest, src, os_on);
        }
        else {
            printf("mswitch apply off to %p\n", dest);
            apply_req(dest, src, os_off);
        }
        break;
    case ot_teleporter:
        break;
    default:
        assert(0);
        break;
    }
}


void apply_all_links(void){
    int i, j, k;
    obj_t *src, *dest;
    int link_no;
    link_table_entry_t link_table[LINK_NO_MAX] = {{0, 0, {NULL}, {NULL}}};
    link_table_entry_t *entry;

    // link_src/dest_entriesをすべてテーブルに登録する
    for (i = 0 ; i < link_src_entry_num ; i++){
        src = link_src_entries[i].obj;
        link_no = link_src_entries[i].link_no;
        entry = &link_table[link_no];
        if (entry->src_num >= MULTI_LINK_MAX){
            log_error(" too many objects using link label '%c' for output (limit %d).\n",
                      LinkNo2Chr(link_no), MULTI_LINK_MAX);
        }
        entry->srcs[entry->src_num] = src;
        entry->src_num ++;
    }
    for (i = 0 ; i < link_dest_entry_num ; i++){
        dest = link_dest_entries[i].obj;
        link_no = link_dest_entries[i].link_no;
        entry = &link_table[link_no];
        if (entry->dest_num >= MULTI_LINK_MAX){
            log_error("too many objects using link label '%c' for input (limit %d).\n",
                      LinkNo2Chr(link_no), MULTI_LINK_MAX);
        }
        entry->dests[entry->dest_num] = dest;
        entry->dest_num ++;
	}
    // 重複検査しつつ接続
    for (i = 0 ; i < LINK_NO_MAX ; i++){
        entry = &link_table[i];
        if (entry->src_num == 0 && entry->dest_num == 0){
            // src, destともに空なら未使用ラベルなのでなにもしない
            //            printf("label %c not used.\n", LinkNo2Chr(i));
        }
        // src, destの一方が空なら警告
        else if (entry->src_num > 0 && entry->dest_num == 0){
            log_warn("no destination for link label '%c'.\n", LinkNo2Chr(i));
        }
        else if (entry->src_num == 0 && entry->dest_num > 0){
            log_warn("no source for link label '%c'.\n", LinkNo2Chr(i));
        }
        // src, destとも1つなら接続処理
        else if (entry->src_num == 1 && entry->dest_num == 1){
            src = entry->srcs[0];
            dest = entry->dests[0];
            if (src->type == ot_hub){
                register_hub_out(src, dest);                
            }
            else {
                src->link = dest;
            }
            if (dest->type == ot_hub){
                register_hub_in(dest, src);
            }
            else {
                // ハブでないdest側オブジェクトはなにもしなくていい
            }
        }
        // 一対多、多対一、多対多
        else {
            // destが複数の場合、srcがハブでなければならない
            if (entry->dest_num > 1){
                for (j = 0 ; j < entry->src_num ; j++){
                    if (entry->srcs[j]->type != ot_hub){
                        log_error("multiple destination for link '%c'.\n", 
                                  LinkNo2Chr(i));
                    }
                }
            }
            for (j = 0 ; j < entry->src_num ; j++){
                for (k = 0 ; k < entry->dest_num ; k++){
                    src = entry->srcs[j];
                    dest = entry->dests[k];
                    if (src->type == ot_hub){
                        register_hub_out(src, dest);                
                    }
                    // @note 上のチェックにより、srcがハブでないならdestは単数のはず
                    else {
                        src->link = dest;
                    }
                    if (dest->type == ot_hub){
                        register_hub_in(dest, src);
                    }
                    else {
                        // ハブでないdest側オブジェクトはなにもしなくていい
                    }
                }
            }
        }
    }
    printf("----- HUB LIST -----\n");
	for (i = 0 ; i < link_src_entry_num ; i++){
        src = link_src_entries[i].obj;
        if (src->type == ot_hub){
            printf("src %d ", i);
            print_hub(src);
        }
    }
    for (i = 0 ; i < link_dest_entry_num ; i++){
		dest = link_dest_entries[i].obj;
        if (dest->type == ot_hub){
            printf("dest %d ", i);
            print_hub(dest);
        }
    }
    printf("----- HUB LIST END.-----\n");
    // オブジェクトの初期状態を伝搬させる
    // ※setupでの伝搬ループだけでよさそうだが、
    //   ループ内でスイッチの変化が起きなければ 
    //   (i.e. スイッチがリンク連鎖の開始点なら)
    //   伝搬すべきreqが発生しない。
    //   各setup_xxx内で強制的に変化を起こす方が綺麗かもしれない。
    // !!! set linkループが終わってから改めて初期伝搬用のループを
    //     回す方が正しい気がするが、そのように修正すると正しく動かない
    //    }
    for (i = 0 ; i < link_src_entry_num ; i++){
        src = link_src_entries[i].obj;
        if (src->type == ot_hub){
            // ハブは起点にならないからここでは無視してよいはず
        }
        else {
            apply_init_req(src->link, src);
        }
    }

}

void clear_links(void){
    int i;

    link_src_entry_num = 0;
    link_dest_entry_num = 0;
    for (i = 0 ; i < LINK_NO_MAX ; i++){
		link_src_entries[i].obj = NULL;
		link_dest_entries[i].obj = NULL;
    }
}

void setup_stage(void){
    int x, rx, y, ry, len;
    map_options_t options;

    dir_t dir = di_none;
    obj_t *obj;
    obj_type_t type;
    int link_no;
    onoff_t onoff;
	logic_t logic;

    setup_obj_list();
	set_chee(NULL);
    current_stage.gondin_total = 0;
	current_stage.sign_no = INVALID_LINK_NO;
	clear_emap();
    clear_links();

    for (y = 0 ; y < MAP_GRID_HEIGHT ; y++){
        for (x = 0 ; x < MAP_GRID_WIDTH ; x++){
            type = symbol2type(GetMapSymbol(x, y));
            //                        printf("(%2d, %2d) [%c](%d) %d\n", x, y, GetMapSymbol(x, y), GetMapSymbol(x, y), type);
            if (type == 0){
                log_error("Undefined character '%c'(%d) appears at (%d, %d)"
                          " in stage '%s'.\n", 
                          GetMapSymbol(x, y),
                          GetMapSymbol(x, y),
                          x, y,
                          current_stage.name);
                exit(1);
            }
            if (type != ot_blank &&
                (obj_introduction[type] < 0 ||  
                 current_stage.no < obj_introduction[type]))
            {
                log_error("Object '%s'(%d) appears at (%d, %d) "
                          "before introduction in stage %d '%s'"
                          " (introduction in stage %d).\n", 
                          obj_type2name(type), type,
                          x, y,
                          current_stage.no, current_stage.name, 
                          obj_introduction[type]);
                exit(1);
            }
            switch(type){
			case ot_blank:
				break;
            case ot_chee:
                if (get_chee()){
					log_error("Multiple CheeChee in stage '%s'.\n", current_stage.name);
                    exit(1);
                }
                get_obj_options(&current_stage, x, y, type, &options);
                dir = (options.dir == di_none) ? di_left : options.dir;
                create_obj_grid(type, x, y, dir);
                break;
            case ot_hatopo:
			case ot_yaramasi:
			case ot_ojigiri:
            case ot_sue:
            case ot_chu:
            case ot_mingo:
            case ot_jumeme:
                get_obj_options(&current_stage, x, y, type, &options);
                dir = (options.dir == di_none) ? di_left : options.dir;
                create_obj_grid(type, x, y, dir);
                break;
            case ot_nya:
                get_obj_options(&current_stage, x, y, type, &options);
                dir = (options.dir == di_none) ? di_left : options.dir;
                obj = create_obj_grid(type, x, y, dir);
                link_no = options.link_no1;
                if (link_no != INVALID_LINK_NO){
                    register_link_dest(obj, link_no);
                }
                break;
            case ot_wall:
            case ot_fwall:
            case ot_iwall:
			case ot_goal:
            case ot_slope_l:
            case ot_slope_r:
            case ot_stairs_l:
            case ot_stairs_r:
            case ot_rstairs_l:
            case ot_rstairs_r:
            case ot_gagga:
                create_obj_grid(type, x, y, di_none);
                break;
            case ot_gondin:
                current_stage.gondin_total ++;
                create_obj_grid(type, x, y, di_none);
                break;
            case ot_sting:
                get_obj_options(&current_stage, x, y, type, &options);
                dir = (options.dir == di_none) ? di_left : options.dir;
                create_obj_grid(type, x, y, dir);
                break;
            case ot_pipe:	
                get_obj_options(&current_stage, x, y, type, &options);
                dir = (options.dir == di_none) ? di_up : options.dir;
                link_no = options.link_no1;
                obj = create_obj_grid(type, x, y, dir);
                if (link_no != INVALID_LINK_NO){
                    register_link_dest(obj, link_no);
                }
                break;
			case ot_sign:
			case ot_sign_hang:
                get_obj_options(&current_stage, x, y, type, &options);
                create_sign_grid(type, x, y, options.link_no1);
				break;
            case ot_lift:
                get_obj_options(&current_stage, x, y, type, &options);
                dir = (options.dir == di_none) ? di_left : options.dir;
                link_no = options.link_no1;
                obj = create_obj_grid(ot_lift, x, y, dir);
                if (link_no != INVALID_LINK_NO){
                    register_link_dest(obj, link_no);
                }
                len = 1;
                rx = x+1;
                while (symbol2type(GetMapSymbol(rx, y)) == ot_lift){
                    len++;
                    rx++;
                }
                set_grid_width_dynamic(obj, len);
                x = rx-1;
                break;
            case ot_barrier:
                // すぐ上もバリアなら、すでに処理済みのバリアの一部なのでパスする
                if (symbol2type(GetMapSymbol(x, y-1)) == ot_barrier) break;
                get_obj_options(&current_stage, x, y, type, &options);
                link_no = options.link_no1;
                obj = create_obj_grid(ot_barrier, x, y, dir);
                if (link_no != INVALID_LINK_NO){
                    register_link_dest(obj, link_no);
                }
                len = 1;
                rx = x+1;
                while (symbol2type(GetMapSymbol(rx, y)) == ot_barrier){
                    len++;
                    rx++;
                }
                set_grid_width_dynamic(obj, len);
                x = rx-1;
                len = 1;
                ry = y+1;
                while (symbol2type(GetMapSymbol(x, ry)) == ot_barrier){
                    len++;
                    ry++;
                }
                set_grid_height_dynamic(obj, len);
                break;
            case ot_teleporter:
                get_obj_options(&current_stage, x, y, type, &options);
                obj = create_obj_grid(type, x, y, di_none);
                link_no = options.link_no1;
                if (link_no != INVALID_LINK_NO){
                    register_link_src(obj, link_no);
                    link_no = options.link_no2;
                    if (link_no != INVALID_LINK_NO){
                        register_link_dest(obj, link_no);
                    }
                }
                len = 1;
                rx = x+1;
                while (symbol2type(GetMapSymbol(rx, y)) == type){
                    len++;
                    rx++;
                }
                set_grid_width_static(obj, len);
                x = rx-1;
                break;
            case ot_teleport_dest:
                get_obj_options(&current_stage, x, y, type, &options);
                obj = create_obj_grid(type, x, y, di_none);
                link_no = options.link_no1;
                if (link_no != INVALID_LINK_NO){
                    register_link_dest(obj, link_no);
                }
                break;
            case ot_mswitch:
            case ot_aswitch:
                get_obj_options(&current_stage, x, y, type, &options);
                obj = create_obj_grid(type, x, y, di_none);
                link_no = options.link_no1;
                if (link_no != INVALID_LINK_NO){
                    register_link_src(obj, link_no);
                }
                onoff = options.onoff;
                if (onoff == oo_on){
                    if (type == ot_aswitch){
//						on_aswitch(obj);
                        SetState(obj, os_req_on);
                    }
                    else {
                        obj->u.mswitch.is_reverse = 1;
						SetState(obj, os_on);
                        obj->motion = ii_mswitch_on;
                    }
                }
                len = 1;
                rx = x+1;
                while (symbol2type(GetMapSymbol(rx, y)) == type){
                    len++;
                    rx++;
                }
                if (type == ot_mswitch){
                    set_grid_width_static(obj, len);
                }
                else {
                    set_grid_width_dynamic(obj, len);
                }
                x = rx-1;
                break;
            case ot_hub:
                get_obj_options(&current_stage, x, y, type, &options);
                obj = create_obj_grid(type, x, y, di_none);
                link_no = options.link_no1;
                if (link_no != INVALID_LINK_NO){
                    register_link_src(obj, link_no);
                    link_no = options.link_no2;
                    if (link_no != INVALID_LINK_NO){
                        register_link_dest(obj, link_no);
                    }
                }
                logic = options.logic;
				if (logic != lo_invalid){
					set_hub_logic(obj, logic);
				}
                onoff = options.onoff;
                if (onoff == oo_off){
                    set_hub_reverse(obj, 1);
                }
                break;
            default:
                break;
            }
        }
    }
    if (get_chee() == NULL){
        log_error("No CheeChee in stage '%s'.\n", current_stage.name);
        exit(1);
    }

    apply_all_links();
    setup_emap();
    setup_stage_obj_list();
}

// 行末の改行を取り除く
void chop_line(char *tline){
    int last = strlen(tline)-1;
    if (tline[last] == '\n'){
        tline[last] = '\0';
    }
}


void update_stage_record(game_t *game){
	stage_thumb_t *thumb;
    int i;

    //	print_stage_record();
    for (i = re_begin ; i < re_end ; i++){
        if (current_stage_record.max[i] < game->record[i]){
            current_stage_record.max[i] = game->record[i];
        }
        if (current_stage_record.min[i] > game->record[i]){
            current_stage_record.min[i] = game->record[i];
        }
        current_stage_record.sum[i] += game->record[i];
    }
    //	print_stage_record();
    thumb = get_current_thumb();
	thumb->top_score = current_stage_record.max[re_score];
	thumb->clear_num= current_stage_record.max[re_clear];
	set_thumb_stars(thumb);
    //    update_thumbs_info();
}


int load_stage_record(char *name, record_t *stage_record, unsigned char stage_major_version, unsigned char stage_minor_version){
    FILE *fp;
	char path[BUF_SIZE];
    unsigned char file_major_version = 0, file_minor_version = 0;
	unsigned char major_version = MAJOR_VERSION, minor_version = MINOR_VERSION;
	unsigned char file_stage_major_version = 0, file_stage_minor_version = 0;

    printf("stage_version:Ver%d.%02d\n", stage_major_version, stage_minor_version);

	sprintf(path, "%s/%s.%s", STAGE_RECORD_DIR, name, STAGE_RECORD_EXT);
	if ((fp = open_data(path, "rb",
                        APPLICATION_ID, APPLICATION_ID_SIZE,
                        &file_major_version, &file_minor_version)) == NULL)
    {
		// 未プレイならファイルがないのが当然なのでいちいち警告しない
		//        log_warn("Cannot open stage record file %s\n", path);
        return 0;
    }
    fread(&file_stage_major_version, sizeof(unsigned char), 1, fp);
    fread(&file_stage_minor_version, sizeof(unsigned char), 1, fp);
    // 旧バージョンなら読み込み中止
    if (GetVersion(file_major_version, file_minor_version) <
        GetVersion(major_version, minor_version))
    {
        log_warn("Cannot read record file '%s' of older game(Ver%d.%02d)... renamed.\n",
                 path, file_major_version, file_minor_version);
        fclose(fp);
        rename_data(path);
        return 0;
    }
    if (GetVersion(file_stage_major_version, file_stage_minor_version) <
        GetVersion(stage_major_version, stage_minor_version))
    {
        log_warn("Cannot read record file '%s' of older stage"
                 "(Ver%d.%02d. Currently Ver%d.%02d)... renamed.\n",
                 path, 
                 file_stage_major_version, file_stage_minor_version,
                 stage_major_version, stage_minor_version);
        fclose(fp);
        rename_data(path);
        return 0;
    }
    //	printf("load %d mins\n", re_end);
    fread_uint(fp, stage_record->min, re_end);
    //	printf("loaded min:%d %d %d\n", stage_record->min[0],stage_record->min[1],stage_record->min[2]); 
    //	printf("load %d maxs\n", re_end);
    fread_uint(fp, stage_record->max, re_end);
    //	printf("load %d sums\n", re_end);
    fread_uint(fp, stage_record->sum, re_end);
    fclose(fp);
    return 1;
}

void load_current_stage_record(char *name, unsigned char stage_major_version, unsigned char stage_minor_version)
{
    int i;
    //	if (!load_stage_record(name, &current_stage_record, current_stage.major_version, current_stage.minor_version)){
	if (!load_stage_record(name, &current_stage_record, 
                           stage_major_version, stage_minor_version))
    {
		memset(&current_stage_record, 0, sizeof(current_stage_record));
        for (i = 0 ; i < re_end ; i++){
            current_stage_record.min[i] = MINMAX_MAX;
        }
	}
    //	print_stage_record();
}


int save_stage_record(char *name, record_t *stage_record, unsigned char stage_major_version, unsigned char stage_minor_version){
	FILE *fp;
	char path[BUF_SIZE];

	sprintf(path, "%s/%s", get_save_dpath(), STAGE_RECORD_DIR);
// 	STAGE_RECORD_DIRがなければ作る
    if (check_dir(path) == 0){
        log_warn("Cannot find '%s'.\n", path);
        if (make_dir(path) == 0){
            log_warn("Created '%s'.\n", path);
        }
        else {
            log_warn("Creating '%s' failed.\n", path);
            return 0;
        }
	}
	sprintf(path, "%s/%s.%s", STAGE_RECORD_DIR, name, STAGE_RECORD_EXT);
    if ((fp = open_data(path, "wb",
                        APPLICATION_ID, APPLICATION_ID_SIZE,
                        &major_version, &minor_version)) == NULL)
    {
		log_warn("Failed saving stage record %s.\n", path);
        return 0;
    }
    fwrite(&stage_major_version, sizeof(unsigned char), 1, fp);
    fwrite(&stage_minor_version, sizeof(unsigned char), 1, fp);
    fwrite_uint(fp, stage_record->min, re_end);
	printf("saved min:%d %d %d\n", stage_record->min[0],stage_record->min[1],stage_record->min[2]); 
    fwrite_uint(fp, stage_record->max, re_end);
    fwrite_uint(fp, stage_record->sum, re_end);
    /*
    fwrite_uint(fp, &current_stage_record.top_score, 1);
	fwrite_uint(fp, &current_stage_record.top_time, 1);
	fwrite(&current_stage_record.top_gondin_eaten, 1, 1, fp);
    */
	fclose(fp);
	return 1;
}

void save_current_stage_record(char *name){
	save_stage_record(name, &current_stage_record, current_stage.major_version, current_stage.minor_version);
}

unsigned int get_current_stage_record_min(record_entry_t entry){
    return current_stage_record.min[entry];
}
unsigned int get_current_stage_record_max(record_entry_t entry){
    return current_stage_record.max[entry];
}
unsigned int get_current_stage_record_sum(record_entry_t entry){
    return current_stage_record.sum[entry];
}

void inc_current_stage_record(record_entry_t entry){
    current_stage_record.sum[entry] ++;
}

objective_t *get_stage_objective(int no){
    return &stage_objectives[no];
}

/*
font_id_t objective_id2font(int id){
    return objective_infos[id].font;
}
*/

#ifdef LOCK_STAGE
int get_introduction_stage(obj_type_t type){
    return obj_introduction[type];
}
#endif

#ifdef DUMMY
void print_stage_record(void){
    printf("score:%d-%d %d\n",
           CurrentStageRecordMin(re_score), 
           CurrentStageRecordMax(re_score),
           CurrentStageRecordSum(re_score));
    printf("time:%d-%d %d\n",
           CurrentStageRecordMin(re_time), 
           CurrentStageRecordMax(re_time),
           CurrentStageRecordSum(re_time));
    printf("time:%d-%d %d\n",
           CurrentStageRecordMin(re_gondin_eaten), 
           CurrentStageRecordMax(re_gondin_eaten),
           CurrentStageRecordSum(re_gondin_eaten));
}
#endif



